import logging
import os
import re

from gixy.core.regexp import Regexp
from gixy.core.variable import Variable

LOG = logging.getLogger(__name__)

BUILTIN_VARIABLES = {
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_uri
    "uri": r"/[^\x20\t]*",
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_document_uri
    "document_uri": r"/[^\x20\t]*",
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_arg_
    "arg_": r"[^\s&]+",
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_args
    "args": r"[^\s]+",
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_query_string
    "query_string": r"[^\s]+",
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_request_uri
    "request_uri": r"/[^\s]*",
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_http_
    "http_": r"[\x21-\x7e]",
    # https://nginx.org/en/docs/http/ngx_http_upstream_module.html#var_upstream_http_
    "upstream_http_": "",
    # https://nginx.org/en/docs/http/ngx_http_upstream_module.html#var_upstream_cookie_
    "upstream_cookie_": "",
    # https://nginx.org/en/docs/http/ngx_http_proxy_module.html#var_proxy_add_x_forwarded_for
    "proxy_add_x_forwarded_for": "",
    # https://nginx.org/en/docs/http/ngx_http_proxy_module.html#var_proxy_host
    "proxy_host": "",
    # https://nginx.org/en/docs/http/ngx_http_proxy_module.html#var_proxy_port
    "proxy_port": "",
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_proxy_protocol_addr
    # https://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_proxy_protocol_addr
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_proxy_protocol_port
    # https://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_proxy_protocol_port
    "proxy_protocol_port": "",
    # https://nginx.org/en/docs/http/ngx_http_fastcgi_module.html#var_fastcgi_path_info
    "fastcgi_path_info": "",
    # https://nginx.org/en/docs/http/ngx_http_fastcgi_module.html#var_fastcgi_script_name
    "fastcgi_script_name": "",
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_content_type
    "content_type": "",
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_cookie_
    "cookie_": "",
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_host
    "host": "",
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_hostname
    # https://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_hostname
    "hostname": "",
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_limit_rate
    "limit_rate": "",
    # https://nginx.org/en/docs/http/ngx_http_memcached_module.html#var_memcached_key
    "memcached_key": "",
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_realpath_root
    "realpath_root": "",
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_remote_user
    "remote_user": "",
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_request
    "request": "",
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_request_body
    "request_body": "",
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_request_completion
    "request_completion": "",
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_request_filename
    "request_filename": "",
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_request_id
    "request_id": "",
    # https://nginx.org/en/docs/http/ngx_http_slice_module.html#var_slice_range
    "slice_range": "",
    # https://nginx.org/en/docs/http/ngx_http_secure_link_module.html#var_secure_link
    "secure_link": "",
    # https://nginx.org/en/docs/http/ngx_http_secure_link_module.html#var_secure_link_expires
    "secure_link_expires": "",
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_sent_http_
    "sent_http_": "",
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_server_name
    "server_name": "",
    # "Secure" variables that can't content or strictly limited user input
    # https://nginx.org/en/docs/http/ngx_http_browser_module.html#var_ancient_browser
    "ancient_browser": None,
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_binary_remote_addr
    # https://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_binary_remote_addr
    "binary_remote_addr": None,
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_body_bytes_sent
    "body_bytes_sent": None,
    # https://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_bytes_received
    "bytes_received": None,
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_bytes_sent
    # https://nginx.org/en/docs/http/ngx_http_log_module.html#var_bytes_sent
    # https://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_bytes_sent
    "bytes_sent": None,
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_connection
    # https://nginx.org/en/docs/http/ngx_http_log_module.html#var_connection
    # https://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_connection
    "connection": None,
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_connection_requests
    # https://nginx.org/en/docs/http/ngx_http_log_module.html#var_connection_requests
    "connection_requests": None,
    # https://nginx.org/en/docs/http/ngx_http_stub_status_module.html#var_connections_active
    "connections_active": None,
    # https://nginx.org/en/docs/http/ngx_http_stub_status_module.html#var_connections_reading
    "connections_reading": None,
    # https://nginx.org/en/docs/http/ngx_http_stub_status_module.html#var_connections_waiting
    "connections_waiting": None,
    # https://nginx.org/en/docs/http/ngx_http_stub_status_module.html#var_connections_writing
    "connections_writing": None,
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_content_length
    "content_length": None,
    # https://nginx.org/en/docs/http/ngx_http_ssi_module.html#var_date_gmt
    "date_gmt": None,
    # https://nginx.org/en/docs/http/ngx_http_ssi_module.html#var_date_local
    "date_local": None,
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_document_root
    "document_root": "/etc/nginx",
    # https://nginx.org/en/docs/http/ngx_http_geoip_module.html
    # https://nginx.org/en/docs/stream/ngx_stream_geoip_module.html
    "geoip_": None,
    # https://nginx.org/en/docs/http/ngx_http_gzip_module.html#var_gzip_ratio
    "gzip_ratio": None,
    # https://nginx.org/en/docs/http/ngx_http_v2_module.html#var_http2
    "http2": None,
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_https
    "https": None,
    # https://nginx.org/en/docs/http/ngx_http_referer_module.html#var_invalid_referer
    "invalid_referer": None,
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_is_args
    "is_args": None,
    # https://nginx.org/en/docs/http/ngx_http_auth_jwt_module.html
    "jwt_": None,
    # https://nginx.org/en/docs/http/ngx_http_browser_module.html#var_modern_browser
    "modern_browser": None,
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_msec
    # https://nginx.org/en/docs/http/ngx_http_log_module.html#var_msec
    # https://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_msec
    "msec": None,
    # https://nginx.org/en/docs/http/ngx_http_browser_module.html#var_msie
    "msie": None,
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_nginx_version
    # https://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_nginx_version
    "nginx_version": None,
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_pid
    # https://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_pid
    "pid": None,
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_pipe
    # https://nginx.org/en/docs/http/ngx_http_log_module.html#var_pipe
    "pipe": None,
    # https://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_protocol
    "protocol": None,
    # https://nginx.org/en/docs/http/ngx_http_realip_module.html#var_realip_remote_addr
    # https://nginx.org/en/docs/stream/ngx_stream_realip_module.html#var_realip_remote_addr
    # https://nginx.org/en/docs/http/ngx_http_realip_module.html#var_realip_remote_port
    # https://nginx.org/en/docs/stream/ngx_stream_realip_module.html#var_realip_remote_port
    "realip_remote_port": None,
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_remote_addr
    # https://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_remote_addr
    "remote_addr": None,
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_remote_port
    # https://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_remote_port
    "remote_port": None,
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_request_body_file
    "request_body_file": None,
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_request_length
    # https://nginx.org/en/docs/http/ngx_http_log_module.html#var_request_length
    "request_length": None,
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_request_method
    "request_method": None,
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_request_time
    # https://nginx.org/en/docs/http/ngx_http_log_module.html#var_request_time
    "request_time": None,
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_scheme
    "scheme": None,
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_server_addr
    # https://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_server_addr
    "server_addr": None,
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_server_port
    # https://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_server_port
    "server_port": None,
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_server_protocol
    "server_protocol": None,
    # https://nginx.org/en/docs/http/ngx_http_session_log_module.html#var_session_log_binary_id
    "session_log_binary_id": None,
    # https://nginx.org/en/docs/http/ngx_http_session_log_module.html#var_session_log_id
    "session_log_id": None,
    # https://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_session_time
    "session_time": None,
    # https://nginx.org/en/docs/http/ngx_http_spdy_module.html#var_spdy
    "spdy": None,
    # https://nginx.org/en/docs/http/ngx_http_spdy_module.html#var_spdy_request_priority
    "spdy_request_priority": None,
    # https://nginx.org/en/docs/http/ngx_http_ssl_module.html
    # https://nginx.org/en/docs/stream/ngx_stream_ssl_module.html
    "ssl_": None,
    # https://nginx.org/en/docs/http/ngx_http_core_module.html#var_status
    # https://nginx.org/en/docs/http/ngx_http_log_module.html#var_status
    # https://nginx.org/en/docs/stream/ngx_stream_core_module.html#var_status
    "status": None,
    # https://nginx.org/en/docs/http/ngx_http_core_module.html
    "tcpinfo_": None,
    # https://nginx.org/en/docs/http/ngx_http_core_module.html
    # https://nginx.org/en/docs/http/ngx_http_log_module.html
    # https://nginx.org/en/docs/stream/ngx_stream_core_module.html
    "time_iso8601": None,
    # https://nginx.org/en/docs/http/ngx_http_core_module.html
    # https://nginx.org/en/docs/http/ngx_http_log_module.html
    # https://nginx.org/en/docs/stream/ngx_stream_core_module.html
    "time_local": None,
    # https://nginx.org/en/docs/http/ngx_http_userid_module.html#var_uid_got
    "uid_got": None,
    # https://nginx.org/en/docs/http/ngx_http_userid_module.html#var_uid_reset
    "uid_reset": None,
    # https://nginx.org/en/docs/http/ngx_http_userid_module.html#var_uid_set
    "uid_set": None,
    # https://nginx.org/en/docs/http/ngx_http_upstream_module.html#var_upstream_addr
    # https://nginx.org/en/docs/stream/ngx_stream_upstream_module.html#var_upstream_addr
    "upstream_addr": None,
    # https://nginx.org/en/docs/http/ngx_http_upstream_module.html#var_upstream_bytes_received
    # https://nginx.org/en/docs/stream/ngx_stream_upstream_module.html#var_upstream_bytes_received
    "upstream_bytes_received": None,
    # https://nginx.org/en/docs/stream/ngx_stream_upstream_module.html#var_upstream_bytes_sent
    "upstream_bytes_sent": None,
    # https://nginx.org/en/docs/http/ngx_http_upstream_module.html#var_upstream_cache_status
    "upstream_cache_status": None,
    # https://nginx.org/en/docs/http/ngx_http_upstream_module.html#var_upstream_connect_time
    # https://nginx.org/en/docs/stream/ngx_stream_upstream_module.html#var_upstream_connect_time
    "upstream_connect_time": None,
    # https://nginx.org/en/docs/stream/ngx_stream_upstream_module.html#var_upstream_first_byte_time
    "upstream_first_byte_time": None,
    # https://nginx.org/en/docs/http/ngx_http_upstream_module.html#var_upstream_header_time
    "upstream_header_time": None,
    # https://nginx.org/en/docs/http/ngx_http_upstream_module.html#var_upstream_response_length
    "upstream_response_length": None,
    # https://nginx.org/en/docs/http/ngx_http_upstream_module.html#var_upstream_response_time
    "upstream_response_time": None,
    # https://nginx.org/en/docs/stream/ngx_stream_upstream_module.html#var_upstream_session_time
    "upstream_session_time": None,
    # https://nginx.org/en/docs/http/ngx_http_upstream_module.html#var_upstream_status
    "upstream_status": None,
}


# Additional variables loaded from drop-in configuration files
EXTRA_VARIABLES = {}


def clear_custom_variables():
    """Clear variables loaded from drop-in files (useful for tests)."""
    EXTRA_VARIABLES.clear()


def _normalize_value_token(token):
    """Parse a token from a drop-in variable file into a usable value.

    Supported forms:
    - "" or '' → empty string (treated as not user-controlled)
    - r'...'/r"..." → regex pattern string
    - '...'/"..." → literal string
    - none/null (case-insensitive) → None
    - empty/missing value → empty string
    Trailing commas are tolerated.
    """
    if token is None:
        return ""

    token = token.strip().rstrip(",").strip()
    if token.lower() in ("none", "null"):
        return None

    # Strip possible raw prefix
    raw_prefix = False
    if len(token) > 2 and token[0] in ("r", "R") and token[1] in ("'", '"'):
        raw_prefix = True
        token = token[1:]

    # Strip quotes
    quoted = False
    if len(token) >= 2 and token[0] == token[-1] and token[0] in ("'", '"'):
        quoted = True
        token = token[1:-1]

    # If it was quoted without a raw prefix, treat as a literal string (not regex).
    # Raw-prefixed quoted values (r"..." / r'...') continue to mean "regex pattern".
    if quoted and not raw_prefix:
        return re.escape(token)

    # Otherwise treat as a regex pattern (backward compatible for unquoted tokens).
    return token


def _parse_dropin_file(file_path):
    """Parse a single drop-in file and return a dict of variables.

    Each non-empty, non-comment line should be of the form:
        name <value>
    where value follows _normalize_value_token rules.
    """
    result = {}
    try:
        with open(file_path, "r") as fh:
            for raw_line in fh:
                line = raw_line.strip()
                if not line or line.startswith("#") or line.startswith(";"):
                    continue

                # Fast-path parse: find first whitespace after a valid identifier
                i = 0
                n = len(line)
                # identifier start
                c0 = line[0]
                if not (c0.isalpha() or c0 == "_"):
                    LOG.info(
                        "Skip malformed custom variable line in %s: %r",
                        file_path,
                        raw_line.rstrip("\n"),
                    )
                    continue
                i = 1
                while i < n and (line[i].isalnum() or line[i] == "_"):
                    i += 1
                name = line[:i]
                # skip spaces between name and value
                j = i
                while j < n and line[j].isspace():
                    j += 1
                value_token = line[j:]
                # Allow optional separator characters between name and value
                # If the line used name = value or name: value, drop the first char
                if value_token[:1] in ("=", ":"):
                    value_token = value_token[1:].strip()
                result[name] = _normalize_value_token(value_token)
    except Exception as e:
        LOG.warning("Failed to load custom variables from %s: %s", file_path, e)
    return result


def load_custom_variables_from_dirs(paths):
    """Load additional variables from provided directories.

    - Reads all files with extensions .cfg or .conf
    - Merges into EXTRA_VARIABLES (later files override earlier ones)
    """
    if not paths:
        return
    for base in paths:
        if not base:
            continue
        expanded = os.path.expanduser(base)
        if not os.path.isdir(expanded):
            continue
        try:
            entries = sorted(os.listdir(expanded))
        except OSError:
            continue
        for fname in entries:
            if not (fname.endswith(".cfg") or fname.endswith(".conf")):
                continue
            fpath = os.path.join(expanded, fname)
            if not os.path.isfile(fpath):
                continue
            parsed = _parse_dropin_file(fpath)
            if parsed:
                EXTRA_VARIABLES.update(parsed)


def _iter_all_variable_items():
    # EXTRA overrides BUILTIN on duplicates
    # Preserve prefix variables (ending with '_') semantics
    # Order: EXTRA first, then BUILTIN
    for k, v in EXTRA_VARIABLES.items():
        yield k, v
    for k, v in BUILTIN_VARIABLES.items():
        if k not in EXTRA_VARIABLES:
            yield k, v


def is_builtin(name):
    if isinstance(name, int):
        # Indexed variables can't be builtin
        return False
    for builtin, _ in _iter_all_variable_items():
        if builtin.endswith("_"):
            if name.startswith(builtin):
                return True
        elif name == builtin:
            return True
    return False


def builtin_var(name):
    for builtin, regexp in _iter_all_variable_items():
        if builtin.endswith("_"):
            if not name.startswith(builtin):
                continue
        elif name != builtin:
            continue

        if regexp:
            return Variable(
                name=name, value=Regexp(regexp, strict=True, case_sensitive=False)
            )
        return Variable(name=name, value="builtin", have_script=False)
    return None


def fake_var(name):
    return Variable(name=name, value=name, have_script=False)



import logging
import os

import gixy
from gixy.core import builtin_variables as builtins
from gixy.core.config import Config
from gixy.core.context import get_context, pop_context, purge_context, push_context
from gixy.core.plugins_manager import PluginsManager
from gixy.directives.directive import MapDirective
from gixy.parser.nginx_parser import NginxParser

LOG = logging.getLogger(__name__)


class Manager(object):
    def __init__(self, config=None):
        self.root = None
        self.config = config or Config()
        self.auditor = PluginsManager(config=self.config)

    def audit(self, file_path, file_data, is_stdin=False):
        LOG.debug("Audit config file: {fname}".format(fname=file_path))
        # Load custom variables if configured
        try:
            vars_dirs = getattr(self.config, "vars_dirs", None)
            if vars_dirs:
                builtins.load_custom_variables_from_dirs(vars_dirs)
        except Exception as e:
            LOG.debug("Custom variables loading failed: %s", e)
        parser = NginxParser(
            cwd=os.path.dirname(file_path) if not is_stdin else "",
            allow_includes=self.config.allow_includes,
        )
        if is_stdin:
            # Route stdin through parse_string for consistent path-based parsing via tempfile
            self.root = parser.parse_string(
                content=file_data.read(), path_info=file_path
            )
        else:
            # Prefer path-based parsing to avoid temporary files
            self.root = parser.parse_file(file_path)

        push_context(self.root)
        self._audit_recursive(self.root.children)
        # Call post_audit hooks after all directives have been processed
        self.auditor.post_audit(self.root)

    @property
    def results(self):
        for plugin in self.auditor.plugins:
            if plugin.issues:
                yield plugin

    @property
    def stats(self):
        stats = dict.fromkeys(gixy.severity.ALL, 0)
        for plugin in self.auditor.plugins:
            base_severity = plugin.severity
            for issue in plugin.issues:
                # TODO(buglloc): encapsulate into Issue class?
                severity = issue.severity if issue.severity else base_severity
                stats[severity] += 1
        return stats

    def _audit_recursive(self, tree):
        for directive in tree:
            self._update_variables(directive)
            self.auditor.audit(directive)
            if directive.is_block:
                if directive.self_context:
                    push_context(directive)
                self._audit_recursive(directive.children)
                if directive.self_context:
                    pop_context()

    def _update_variables(self, directive):
        # TODO(buglloc): finish him!
        if not directive.provide_variables:
            return

        context = get_context()
        for var in directive.variables:
            if var.name == 0 and not isinstance(directive, MapDirective):
                # All regexps must clean indexed variables
                context.clear_index_vars()
            context.add_var(var.name, var)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        purge_context()



import gixy
from gixy.plugins.plugin import Plugin


class proxy_buffering_off(Plugin):
    """
    Insecure example:
        proxy_buffering off;
    """

    summary = "Do not disable `proxy_buffering`."
    severity = gixy.severity.MEDIUM
    description = "Disabling `proxy_buffering` can increase slow-client DoS risk by tying up upstream connections and workers."
    help_url = "https://gixy.io/plugins/proxy_buffering_off/"
    directives = ["proxy_buffering"]

    def audit(self, directive):
        if not directive.args:
            return

        try:
            value = str(directive.args[0]).strip().lower()
        except (TypeError, IndexError):
            return

        if value == "off":
            self.add_issue(
                directive=directive,
                reason="`proxy_buffering` is set to `off`.",
            )



import gixy
from gixy.core.variable import compile_script
from gixy.plugins.plugin import Plugin


class http_splitting(Plugin):
    r"""
    Insecure examples:
        rewrite ^ http://$host$uri;
        return 301 http://$host$uri;
        proxy_set_header "X-Original-Uri" $uri;
        proxy_pass http://upstream$document_uri;

        location ~ /proxy/(a|b)/(\W*)$ {
            set $path $2;
            proxy_pass http://storage/$path;
        }
    """

    summary = "Possible HTTP-Splitting vulnerability."
    severity = gixy.severity.HIGH
    description = (
        'Using variables that can contain "\\n" or "\\r" may lead to http injection.'
    )
    help_url = "https://gixy.io/plugins/http_splitting/"
    directives = ["rewrite", "return", "add_header", "proxy_set_header", "proxy_pass"]

    def audit(self, directive):
        value = _get_value(directive)
        if not value:
            return

        server_side = directive.name.startswith("proxy_")
        for var in compile_script(value):
            char = ""
            if var.can_contain("\n"):
                char = "\\n"
            elif not server_side and var.can_contain("\r"):
                char = "\\r"
            else:
                continue
            reason = "Variable `${var}` can contain `{char}`, which can enable header/response splitting.".format(
                var=var.name, char=char
            )
            self.add_issue(directive=[directive] + var.providers, reason=reason)


def _get_value(directive):
    if directive.name == "proxy_pass" and len(directive.args) >= 1:
        return directive.args[0]
    elif len(directive.args) >= 2:
        return directive.args[1]
    return None



UNSPECIFIED = "UNSPECIFIED"
LOW = "LOW"
MEDIUM = "MEDIUM"
HIGH = "HIGH"
ALL = [UNSPECIFIED, LOW, MEDIUM, HIGH]


def is_acceptable(current_severity, min_severity):
    return ALL.index(current_severity) >= ALL.index(min_severity)



import gixy
from gixy.plugins.plugin import Plugin


class error_log_off(Plugin):
    """
    Insecure example:
        error_log off;
    """

    summary = "error_log set to off."
    severity = gixy.severity.LOW
    description = "The error_log directive should not be set to off. It should be set to a valid file path."
    help_url = "https://gixy.io/plugins/error_log_off/"
    directives = ["error_log"]

    def audit(self, directive):
        if directive.args and directive.args[0].lower() == "off":
            self.add_issue(
                directive=directive,
                reason="Configured `error_log off;` which treats 'off' as a filename.",
            )



import gixy
from gixy.plugins.plugin import Plugin


class resolver_external(Plugin):
    """
    Syntax for the directive: resolver 127.0.0.1 [::1]:5353 valid=30s;
    """

    summary = "Using external DNS nameservers for resolver."
    severity = gixy.severity.MEDIUM
    description = (
        "Using external nameservers allows an attacker to send spoofed DNS replies to poison the resolver cache, "
        "causing NGINX to proxy requests to an arbitrary upstream server."
    )
    help_url = "https://gixy.io/plugins/resolver_external/"
    directives = ["resolver"]

    def audit(self, directive):
        bad_nameservers = directive.get_external_nameservers()
        if bad_nameservers:
            self.add_issue(
                severity=gixy.severity.HIGH,
                directive=directive,
                reason="Resolver uses external DNS servers: `{dns_servers}`.".format(
                    dns_servers=", ".join(bad_nameservers)
                ),
            )



import gixy
from gixy.plugins.plugin import Plugin


class allow_without_deny(Plugin):
    """
    Warn when an 'allow' directive appears in a context without a corresponding
    'deny all;' (or equivalent restriction) in the same context.
    """

    summary = "Allow directives without a deny restriction."
    severity = gixy.severity.HIGH
    description = "Allow directives should typically be paired with a restrictive deny rule (for example, deny all;) in the same context."
    help_url = "https://gixy.io/plugins/allow_without_deny/"
    directives = ["allow"]

    def __init__(self, config):
        super(allow_without_deny, self).__init__(config)
        self._reported_parents = set()

    def audit(self, directive):
        parent = directive.parent
        if not parent:
            return
        if directive.args == ["all"]:
            # for example, "allow all" in a nested location which allows access to otherwise forbidden parent location
            return

        key = id(parent)
        if key in self._reported_parents:
            return
        self._reported_parents.add(key)

        deny_found = False
        for child in parent.children:
            if child.name == "deny":
                deny_found = True
                break

        if not deny_found:
            reason = "No deny rule was found in the same context; add `deny all;` after the `allow` directives."
            self.add_issue(
                directive=[directive] + list(parent.find_recursive("allow")),
                reason=reason,
            )



from __future__ import absolute_import

import gixy
from gixy.directives import block
from gixy.directives.block import GeoBlock, MapBlock


class BaseFormatter(object):
    skip_parents = {block.Root, block.HttpBlock}

    def __init__(self):
        self.reports = {}
        self.stats = dict.fromkeys(gixy.severity.ALL, 0)
        self._config_cache = {}

    def format_reports(self, reports, stats):
        raise NotImplementedError("Formatter must override format_reports function")

    def feed(self, path, manager):
        self._config_cache.clear()
        for severity in gixy.severity.ALL:
            self.stats[severity] += manager.stats[severity]

        self.reports[path] = []
        for result in manager.results:
            report = self._prepare_result(
                manager.root,
                summary=result.summary,
                severity=result.severity,
                description=result.description,
                issues=result.issues,
                plugin=result.name,
                help_url=result.help_url,
            )
            self.reports[path].extend(report)

    def flush(self):
        return self.format_reports(self.reports, self.stats)

    def _prepare_result(
        self, root, issues, severity, summary, description, plugin, help_url
    ):
        result = {}
        for issue in issues:
            report = dict(
                plugin=plugin,
                summary=issue.summary or summary,
                severity=issue.severity or severity,
                description=issue.description or description,
                help_url=issue.help_url or help_url,
                reason=issue.reason or "",
            )
            key = "".join(report.values())
            expanded_directives = []
            if any(
                isinstance(value, (MapBlock, GeoBlock)) for value in issue.directives
            ):
                for value in issue.directives:
                    if isinstance(value, (MapBlock, GeoBlock)):
                        expanded_directives.extend(value.children)
                    else:
                        expanded_directives.append(value)
                issue.directives = expanded_directives
            report["directives"] = issue.directives
            if key in result:
                result[key]["directives"].extend(report["directives"])
            else:
                result[key] = report

        for report in result.values():
            if report["directives"]:
                config = self._resolve_config(root, report["directives"])
                # Extract location info from first directive with line/file info
                location = self._extract_location(report["directives"])
            else:
                config = ""
                location = None

            del report["directives"]
            report["config"] = config
            report["location"] = location
            yield report

    def _extract_location(self, directives):
        """Extract file and line info from directives for display."""
        for directive in directives:
            if hasattr(directive, "line") and directive.line is not None:
                return {
                    "file": getattr(directive, "file", None),
                    "line": directive.line,
                }
        return None

    def _resolve_config(self, root, directives):
        points = set()
        for directive in directives:
            points.add(directive)
            points.update(p for p in directive.parents)

        cache_key = frozenset(points)
        cached = self._config_cache.get(cache_key)
        if cached is not None:
            return cached

        rendered = "\n".join(self._traverse_tree(root, points, 0))
        self._config_cache[cache_key] = rendered
        return rendered

    def _traverse_tree(self, tree, points, level):
        result = []
        for leap in tree.children:
            if leap not in points:
                continue
            printable = type(leap) not in self.skip_parents
            # Special hack for includes
            # TODO(buglloc): fix me
            have_parentheses = type(leap) != block.IncludeBlock

            if printable:
                if leap.is_block and result:
                    result.append("")
                directive = str(leap).replace("\n", "\n" + "\t" * (level + 1))
                result.append(
                    "{indent:s}{dir:s}".format(indent="\t" * level, dir=directive)
                )

            if leap.is_block:
                result.extend(
                    self._traverse_tree(leap, points, level + 1 if printable else level)
                )
                if printable and have_parentheses:
                    result.append("{indent:s}}}".format(indent="\t" * level))

        if len(result) >= 2 and result[-1].strip("\t") == "}":
            suffix = "\n" + ("\t" * (level + 1))
            if result[-2].endswith(suffix):
                result[-2] = result[-2][:-len(suffix)]

        if result and level == 0 and result[0] != "":
            result = [""] + result

        return result



from __future__ import absolute_import

from gixy.formatters._jinja import load_template
from gixy.formatters.base import BaseFormatter


class ConsoleFormatter(BaseFormatter):
    def __init__(self):
        super(ConsoleFormatter, self).__init__()
        self.template = load_template("console.j2")

    def format_reports(self, reports, stats):
        return self.template.render(reports=reports, stats=stats)



try:
    from cached_property import cached_property
except ImportError:
    from functools import cached_property

from gixy.core.regexp import Regexp
from gixy.core.variable import Variable, compile_script
from gixy.directives.directive import Directive, MapDirective


def get_overrides():
    result = {}
    for klass in Block.__subclasses__():
        if not klass.nginx_name:
            continue

        if not klass.__name__.endswith("Block"):
            continue

        result[klass.nginx_name] = klass
    return result


class Block(Directive):
    nginx_name = None
    is_block = True
    self_context = True

    def __init__(self, name, args):
        super(Block, self).__init__(name, args)
        self.children = []
        self._children_by_name = {}

    def some(self, name, flat=True):
        """Find first directive with given name"""
        if not flat:
            lst = self._children_by_name.get(name)
            return lst[0] if lst else None

        for child in self.children:
            if child.name == name:
                return child
            if flat and child.is_block and not child.self_context:
                result = child.some(name, flat=flat)
                if result:
                    return result
        return None

    def find(self, name, flat=False):
        if not flat:
            return list(self._children_by_name.get(name, ()))

        result = []
        for child in self.children:
            if child.name == name:
                result.append(child)
            if child.is_block and not child.self_context:
                result.extend(child.find(name, flat=True))
        return result

    def find_recursive(self, name):
        result = []
        for child in self.children:
            if child.name == name:
                result.append(child)
            if child.is_block:
                result.extend(child.find_recursive(name))
        return result

    def append(self, directive):
        directive.set_parent(self)
        self.children.append(directive)
        self._children_by_name.setdefault(directive.name, []).append(directive)

    def find_children_directives(self, name):
        """Find directives below the current scope"""
        for child in self.children:
            if child.name == name:
                yield child
            if child.is_block:
                yield from child.find_children_directives(name)

    def find_all_contexts_of_type(self, context_type):
        """Find all block contexts of a specific type (e.g., 'server', 'location') in this scope"""
        for child in self.children:
            if child.is_block and child.name == context_type:
                yield child
            if child.is_block:
                yield from child.find_all_contexts_of_type(context_type)

    def __str__(self):
        return "{name} {args} {{".format(name=self.name, args=" ".join(self.args))


class Root(Block):
    nginx_name = None

    def __init__(self):
        super(Root, self).__init__(None, [])


class HttpBlock(Block):
    nginx_name = "http"

    def __init__(self, name, args):
        super(HttpBlock, self).__init__(name, args)


class ServerBlock(Block):
    nginx_name = "server"

    def __init__(self, name, args):
        super(ServerBlock, self).__init__(name, args)

    def get_names(self):
        return self.find("server_name")

    def __str__(self):
        server_names = [str(sn) for sn in self.find("server_name")]
        if server_names:
            return "server {{\n{0}\n".format("\n".join(server_names[:2]))
        return "server {"


class LocationBlock(Block):
    nginx_name = "location"
    provide_variables = True

    def __init__(self, name, args):
        super(LocationBlock, self).__init__(name, args)
        if len(args) == 2:
            self.modifier, self.path = args
        else:
            self.modifier = None
            self.path = args[0]

    @property
    def is_internal(self):
        return self.some("internal") is not None

    @property
    def is_regex(self):
        return self.modifier and self.modifier in ("~", "~*")

    @cached_property
    def _regexp(self):
        if not self.is_regex:
            return None
        return Regexp(self.path, case_sensitive=self.modifier == "~")

    @cached_property
    def variables(self):
        regexp = self._regexp
        if not regexp:
            return []
        return [
            Variable(name=name, value=group, boundary=None, provider=self)
            for name, group in regexp.groups.items()
        ]

    def needs_anchor(self):
        regexp = self._regexp
        if not regexp:
            return False
        return regexp.needs_tail_anchor()


class IfBlock(Block):
    nginx_name = "if"
    self_context = False

    def __init__(self, name, args):
        super(IfBlock, self).__init__(name, args)
        self.operand = None
        self.value = None
        self.variable = None

        if len(args) == 1:
            # if ($slow)
            self.variable = args[
                0
            ]  # Do not lstrip because this is not used for defining variables
        elif len(args) == 2:
            # if (!-e $foo)
            self.operand, self.value = args
        elif len(args) == 3:
            # if ($request_method = POST)
            self.variable, self.operand, self.value = args
        else:
            raise Exception('Unknown "if" definition, args: {0!r}'.format(args))

    @property
    def provide_variables(self):
        return self.operand in ("~", "~*", "!~", "!~*") and bool(self.value)

    def __str__(self):
        return "{name} ({args}) {{".format(name=self.name, args=" ".join(self.args))

    @property
    def is_regex(self):
        return self.operand and self.operand in ("~", "~*", "!~", "!~*")

    @cached_property
    def variables(self):
        if not self.is_regex or not self.value:
            return []

        boundary = None
        if self.variable:
            try:
                compiled_script = compile_script(self.variable)
                if len(compiled_script) == 1:
                    boundary = compiled_script[0].regexp
            except (IndexError, AttributeError):
                # compile_script may raise if the context of a variable cannot be found, such as in unit tests.
                pass

        regexp = Regexp(self.value, case_sensitive=self.operand in {"~", "!~"})
        result = []
        for name, group in regexp.groups.items():
            result.append(
                Variable(name=name, value=group, boundary=boundary, provider=self)
            )

        return result


class IncludeBlock(Block):
    nginx_name = "include"
    self_context = False

    def __init__(self, name, args):
        super(IncludeBlock, self).__init__(name, args)
        self.file_path = args[0]

    def __str__(self):
        return "include {0};".format(self.file_path)


class MapBlock(Block):
    """
    map $source $destination { <- this part is the block
        default value; <- this part is the directive, but MapBlock sets variables
        key     value; <- this part is the directive, but MapBlock sets variables
        ~*^re(.*)$ $1; <- this part is the directive, but MapBlock sets variables
    } <- this part is the block
    """

    nginx_name = "map"
    self_context = False
    provide_variables = True

    def __init__(self, name, args):
        super(MapBlock, self).__init__(name, args)
        self.source = args[0]
        self.variable = args[1].lstrip("$")

    def gather_map_directives(self, nodes):
        for node in nodes:
            if isinstance(node, MapDirective):
                yield node
            elif isinstance(node, IncludeBlock):
                yield from self.gather_map_directives(node.children)

    @cached_property
    def variables(self):
        vars = []
        for child in self.gather_map_directives(self.children):
            if not isinstance(child, MapDirective):
                continue  # XXX: Should never happen?
            src_val = child.src_val
            dest_val = child.dest_val

            if not child.is_regex:
                vars.append(
                    Variable(
                        name=src_val,
                        value=dest_val,
                        have_script=False,
                        provider=child,
                        ctx=src_val,
                    )
                )
                continue

            result = []
            for name, group in child.regex.groups.items():
                result.append(
                    Variable(
                        name=name,
                        value=group,
                        provider=child,
                        boundary=None,
                        ctx=src_val,
                    )
                )
                break  # Only need the first result (full expression)
            if len(result) != 1:
                continue
            vars.append(
                Variable(
                    name=src_val,
                    value=result[0].value,  # Value is Regexp()
                    boundary=None,
                    provider=child,
                    have_script=False,
                    ctx=src_val,
                ),
            )

        return [
            Variable(
                name=self.variable,
                value=vars,
                boundary=None,
                provider=self,
                have_script=False,
            )
        ]

    def __str__(self):
        return "{0} {1} ${2} {{".format(self.nginx_name, self.source, self.variable)


class GeoBlock(Block):
    """
    geo [$remote_addr] $geo { <- this part
      default        ZZ;
      include        conf/geo.conf;
      delete         127.0.0.0/16;
      proxy          192.168.100.0/24;
      proxy          2001:0db8::/32;
      key            value;
    } <- this part
    """

    nginx_name = "geo"
    self_context = False
    provide_variables = True

    def __init__(self, name, args):
        super(GeoBlock, self).__init__(name, args)
        if len(args) == 1:  # geo uses $remote_addr as default source of the value
            source = "$remote_addr"
            variable = args[0].lstrip("$")
        else:
            source = args[0]
            variable = args[1].lstrip("$")
        self.source = source
        self.variable = variable

    def gather_geo_directives(self, nodes):
        for node in nodes:
            if isinstance(node, MapDirective):
                yield node
            elif isinstance(node, IncludeBlock):
                yield from self.gather_geo_directives(node.children)

    @cached_property
    def variables(self):
        vars = []
        for child in self.gather_geo_directives(self.children):
            src_val = child.src_val
            dest_val = child.dest_val

            vars.append(
                Variable(
                    name=src_val,
                    value=dest_val,
                    boundary=None,
                    provider=child,
                    have_script=False,
                    ctx=src_val,
                ),
            )
        return [
            Variable(
                name=self.variable,
                value=vars,
                boundary=None,
                provider=self,
                have_script=False,
            )
        ]

    def __str__(self):
        return "{0} {1} ${2} {{".format(self.nginx_name, self.source, self.variable)



from __future__ import absolute_import

import json

from gixy.formatters.base import BaseFormatter


class JsonFormatter(BaseFormatter):
    def format_reports(self, reports, stats):
        result = []
        for path, issues in reports.items():
            for issue in issues:
                entry = {
                    "path": path,
                    "plugin": issue["plugin"],
                    "summary": issue["summary"],
                    "severity": issue["severity"],
                    "description": issue["description"],
                    "reference": issue["help_url"],
                    "reason": issue["reason"],
                    "config": issue["config"].strip("\n"),
                }
                location = issue.get("location")
                if location:
                    entry["line"] = location.get("line")
                    if location.get("file"):
                        entry["file"] = location["file"]
                result.append(entry)

        return json.dumps(result, sort_keys=True, indent=2, separators=(",", ": "))



from __future__ import absolute_import

from jinja2 import Environment, PackageLoader

from gixy.utils.text import to_text


def load_template(name):
    env = Environment(
        loader=PackageLoader("gixy", "formatters/templates"),
        trim_blocks=True,
        lstrip_blocks=True,
        autoescape=False,
    )
    env.filters["to_text"] = to_text_filter
    return env.get_template(name)


def to_text_filter(text):
    try:
        return text.encode("latin1").decode("utf-8")
    except UnicodeEncodeError:
        return to_text(text)



import pathlib
import re

from setuptools import find_packages, setup

# FileNotFoundError is not there in Python 2, define it:
try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError

here = pathlib.Path(__file__).parent

# Extract version from gixy/__init__.py
with open(here / "gixy" / "__init__.py", "r", encoding="utf-8") as fd:
    version = re.search(
        r'^version\s*=\s*[\'"]([^\'"]*)[\'"]',
        fd.read(),
        re.MULTILINE,
    ).group(1)

if not version:
    raise RuntimeError("Cannot find version information")

install_requires = [
    "crossplane>=0.5.8",
    'cached-property>=1.2.0;python_version<"3.8"',
    'argparse>=1.4.0;python_version<"3.2"',
    "Jinja2>=2.8",
    "ConfigArgParse>=0.11.0",
    'tldextract==3.1.2; python_version>="3.6" and python_version<"3.7"',
    'tldextract==4.0.0; python_version>="3.7" and python_version<"3.8"',
    'tldextract>=5.1.2,<5.3.0; python_version>="3.8" and python_version<"3.9"',
    'tldextract>=5.3.0; python_version>="3.9"',
]

tests_requires = [
    "pytest>=7.0.0",
    "pytest-xdist",
]

dev_requires = tests_requires + [
    "coverage>=4.3",
    "flake8>=3.2",
    "tox>=2.7.0",
    "setuptools",
    "twine",
]

long_description = None
readme_path = here / "README.md"
docs_index_path = here / "docs" / "en" / "index.md"

try:
    text = readme_path.read_text(encoding="utf-8")
    single_line = text.strip()
    if single_line.endswith(".md") and "\n" not in single_line:
        candidate = here / single_line
        try:
            long_description = candidate.read_text(encoding="utf-8")
        except FileNotFoundError:
            long_description = text
    else:
        long_description = text
except FileNotFoundError:
    try:
        long_description = docs_index_path.read_text(encoding="utf-8")
    except FileNotFoundError:
        long_description = None

setup(
    name="Gixy-Next",
    version=version,

    description=(
        "Open source NGINX configuration security scanner for "
        "detecting nginx security/performance misconfigurations"
    ),

    long_description=long_description,
    long_description_content_type="text/markdown",

    keywords=[
        "nginx",
        "nginx security",
        "nginx hardening",
        "nginx configuration",
        "nginx config",
        "nginx config scanner",
        "nginx configuration checker",
        "nginx config linter",
        "nginx security scanner",
        "nginx configuration static analyzer",
        "nginx vulnerability scanner",
        "nginx.conf security audit",
        "configuration compliance",
        "configuration security",
        "static analysis",
        "ssrf",
        "http response splitting",
        "host header spoofing",
        "version disclosure",
        "redos",
        "gixy",
        "gixy next",
        "gixy-ng",
        "gixyng",
    ],

    author="Joshua Rogers",
    author_email="gixy@joshua.hu",

    url="https://gixy.io/",
    project_urls={
        "Homepage": "https://gixy.io/",
        "Documentation": "https://gixy.io/",
        "Source": "https://github.com/MegaManSec/gixy-next",
        "Issue Tracker": "https://github.com/MegaManSec/gixy-next/issues",
        "Original Gixy": "https://github.com/yandex/gixy",
    },

    packages=find_packages(exclude=["tests", "tests.*"]),
    include_package_data=True,

    install_requires=install_requires,
    extras_require={
        "tests": tests_requires,
        "dev": dev_requires,
        # Optional: ReDoS checks via external API
        "redos": ["requests>=2.20.0"],
    },

    entry_points={
        "console_scripts": [
            "gixy=gixy.cli.main:main",
            "gixy-next=gixy.cli.main:main",
        ],
    },

    license="MPL-2.0",

    license_files=["LICENSE"],

    classifiers=[
        "Environment :: Console",

        "Intended Audience :: System Administrators",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",

        "Topic :: Security",
        "Topic :: System :: Systems Administration",
        "Topic :: Internet :: WWW/HTTP :: Site Management",
        "Topic :: Software Development :: Quality Assurance",
        "Topic :: Software Development :: Testing",
        "Topic :: Utilities",

        "Operating System :: POSIX :: Linux",

        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.14",
        "Programming Language :: Python :: 3.15",
    ],

    python_requires=">=3.6",
)



import pytest

from gixy.directives.block import *
from gixy.directives.directive import *
from gixy.parser.nginx_parser import NginxParser


def _parse(config):
    return NginxParser(cwd="", allow_includes=False).parse_string(config)


def _parse_via_compat_parse(config):
    """Use the deprecated NginxParser.parse() alias for backward-compat tests."""
    return NginxParser(cwd="", allow_includes=False).parse(config)


@pytest.mark.parametrize('config,expected', zip(
    [
        'access_log syslog:server=127.0.0.1,tag=nginx_sentry toolsformat;',
        'user http;',
        'internal;',
        'set $foo "bar";',
        "set $foo 'bar';",
        'proxy_pass http://unix:/run/sock.socket;',
        'rewrite ^/([a-zA-Z0-9]+)$ /$1/${arg_v}.pb break;'
    ],

    [
        [Directive],
        [Directive],
        [Directive],
        [Directive, SetDirective],
        [Directive, SetDirective],
        [Directive],
        [Directive, RewriteDirective]
    ]
))
def test_directive(config, expected):
    assert_config(config, expected)


@pytest.mark.parametrize('config,expected', zip(
    [
        'if (-f /some) {}',
        'map $uri $avar {}',
        'location / {}'
    ],

    [
        [Directive, Block, IfBlock],
        [Directive, Block, MapBlock],
        [Directive, Block, LocationBlock],
    ]
))
def test_blocks(config, expected):
    assert_config(config, expected)


def test_dump_simple():
    config = '''
# configuration file /etc/nginx/nginx.conf:
http {
    include sites/*.conf;
}

# configuration file /etc/nginx/conf.d/listen:
listen 80;

# configuration file /etc/nginx/sites/default.conf:
server {
    include conf.d/listen;
}
    '''

    tree = _parse(config)
    assert isinstance(tree, Directive)
    assert isinstance(tree, Block)
    assert isinstance(tree, Root)

    assert len(tree.children) == 1
    http = tree.children[0]
    assert isinstance(http, Directive)
    assert isinstance(http, Block)
    assert isinstance(http, HttpBlock)

    # After flattening dump includes, server block is directly under http
    assert len(http.children) == 1
    server = http.children[0]
    assert isinstance(server, Directive)
    assert isinstance(server, Block)
    assert isinstance(server, ServerBlock)

    # listen directive from included file is now flattened under server
    assert len(server.children) == 1
    listen = server.children[0]
    assert isinstance(listen, Directive)
    assert listen.args == ['80']


def test_encoding():
    configs = [
        'bar "\xD1\x82\xD0\xB5\xD1\x81\xD1\x82";'
    ]

    for i, config in enumerate(configs):
        _parse(config)


def test_dump_nested_include_resolves_relative_to_root():
    config = '''
# configuration file /etc/nginx/nginx.conf:
http {
    include sites/a.conf;
}

# configuration file /etc/nginx/sites/a.conf:
server {
    include snippets/shared;
}

# configuration file /etc/nginx/snippets/shared:
add_header X-Test 1;
    '''

    tree = _parse(config)
    assert isinstance(tree, Directive)
    assert isinstance(tree, Block)
    assert isinstance(tree, Root)

    assert len(tree.children) == 1
    http = tree.children[0]
    assert isinstance(http, Directive)
    assert isinstance(http, Block)
    assert isinstance(http, HttpBlock)

    # server is directly under http after flattening
    assert len(http.children) == 1
    server = http.children[0]
    assert isinstance(server, Directive)
    assert isinstance(server, Block)
    assert isinstance(server, ServerBlock)

    # add_header from snippets/shared is flattened under server
    assert len(server.children) == 1
    add_header = server.children[0]
    assert isinstance(add_header, Directive)
    assert add_header.name == 'add_header'
    assert add_header.args == ['X-Test', '1']


def test_dump_sibling_includes_resolve_from_prefix():
    config = '''
# configuration file /etc/nginx/nginx.conf:
http {
    include sites/default.conf;
}

# configuration file /etc/nginx/sites/default.conf:
server {
    include conf.d/listen;
    include conf.d/add_header;
}

# configuration file /etc/nginx/conf.d/listen:
listen 80;

# configuration file /etc/nginx/conf.d/add_header:
add_header X-Foo bar;
    '''

    tree = _parse(config)
    http = tree.children[0]
    server = http.children[0]

    assert len(server.children) == 2
    names = [c.name for c in server.children]
    assert 'listen' in names
    assert 'add_header' in names


@pytest.mark.parametrize("config", [
    'user http;',
    'location / {}',
])
def test_parse_alias_matches_parse_string_for_simple_configs(config):
    tree_direct = _parse(config)
    tree_compat = _parse_via_compat_parse(config)

    # Both APIs should return a Root with a single child of the same directive type.
    assert isinstance(tree_direct, Root)
    assert isinstance(tree_compat, Root)
    assert len(tree_direct.children) == len(tree_compat.children) == 1
    assert type(tree_direct.children[0]) is type(tree_compat.children[0])


def test_parse_alias_handles_dump_same_as_parse_string():
    config = '''
# configuration file /etc/nginx/nginx.conf:
http {
    include sites/*.conf;
}

# configuration file /etc/nginx/conf.d/listen:
listen 80;

# configuration file /etc/nginx/sites/default.conf:
server {
    include conf.d/listen;
}
    '''

    tree_direct = _parse(config)
    tree_compat = _parse_via_compat_parse(config)

    # Both trees should have the same high-level shape for dump parsing.
    assert isinstance(tree_direct, Root)
    assert isinstance(tree_compat, Root)
    assert len(tree_direct.children) == len(tree_compat.children) == 1
    assert isinstance(tree_direct.children[0], HttpBlock)
    assert isinstance(tree_compat.children[0], HttpBlock)

def assert_config(config, expected):
    tree = _parse(config)
    assert isinstance(tree, Directive)
    assert isinstance(tree, Block)
    assert isinstance(tree, Root)

    child = tree.children[0]
    for ex in expected:
        assert isinstance(child, ex)



import gixy
from gixy.plugins.plugin import Plugin


class add_header_multiline(Plugin):
    """
    Insecure example:
        add_header Content-Security-Policy "
        default-src: 'none';
        img-src data: https://mc.yandex.ru https://yastatic.net *.yandex.net https://mc.yandex.${tld} https://mc.yandex.ru;
        font-src data: https://yastatic.net;";
    """

    summary = "Multiline response headers."
    severity = gixy.severity.LOW
    description = "Multiline headers are deprecated (see RFC 7230). Some clients never support them (for example, IE/Edge)."
    help_url = "https://gixy.io/plugins/add_header_multiline/"
    directives = ["add_header", "more_set_headers"]

    def audit(self, directive):
        for header, value in directive.headers.items():
            if value is None:
                continue
            if "\n\x20" in value or "\n\t" in value:
                reason = (
                    "Header value contains an obsolete folded newline (header folding)."
                )
                self.add_issue(directive=directive, reason=reason)
                break
            if "\n" in value:
                reason = "Header value contains a newline; the emitted header may be truncated or invalid."
                self.add_issue(directive=directive, reason=reason)
                break



import re

import gixy
from gixy.core.variable import compile_script
from gixy.plugins.plugin import Plugin


class alias_traversal(Plugin):
    r"""
    Insecure examples:

    location /files {
        alias /home/;
    }

    location ~ /site/(l\.)(.*) {
        alias /lol$1/$2;
    }
    """

    summary = "Path traversal via misconfigured alias."
    severity = gixy.severity.HIGH
    description = "Using alias in a prefixed location that doesn't end with a directory separator could lead to path traversal vulnerability."
    help_url = "https://gixy.io/plugins/alias_traversal/"
    directives = ["alias"]

    def audit(self, directive):
        for location in directive.parents:
            if location.name != "location":
                continue

            if location.modifier in ("~", "~*"):
                self.check_regex_location(directive, location)
            elif not location.modifier or location.modifier == "^~":
                self.check_prefix_location(directive, location)

            return

    def check_prefix_location(self, directive, location):
        if not location.path.endswith("/"):
            severity = (
                gixy.severity.HIGH
                if directive.path.endswith("/")
                else gixy.severity.MEDIUM
            )
            self.report_issue(directive, location, severity)

    def check_regex_location(self, directive, location):
        alias_parts = compile_script(directive.path)

        # "\." -> "."
        location_pattern = re.sub(r"\\(.)", r"\1", location.path)

        search_pos = 0
        prev_part = None

        for part in alias_parts:
            if not getattr(part, "regexp", False):
                prev_part = part
                continue

            capture_group = "(" + str(part.value) + ")"
            group_pos = location_pattern.find(capture_group, search_pos)

            if group_pos < 0:
                prev_part = part
                continue

            search_pos = group_pos

            location_has_slash_before = (
                group_pos == 0 and part.must_startswith("/")
            ) or (
                group_pos > 0
                and (
                    location_pattern[group_pos - 1] == "/" or part.must_startswith("/")
                )
            )

            if prev_part is None:
                self.report_issue(directive, location, gixy.severity.HIGH)
            elif not location_has_slash_before:
                alias_has_slash_before = str(prev_part.value).endswith("/")

                if alias_has_slash_before:
                    # location /bar(.*) ~ { alias /foo/$1; }
                    # only dangerous if the captured part can start with "."
                    if part.can_startswith("."):
                        if part.can_contain("/"):
                            # location /site(.*) ~ { alias /lol/$1; }
                            self.report_issue(directive, location, gixy.severity.HIGH)
                        else:
                            # location /site([^/]*) ~ { alias /lol/$1; }
                            self.report_issue(directive, location, gixy.severity.MEDIUM)
                else:
                    alias_has_slash_before = str(prev_part.value).endswith("/")

                    if not alias_has_slash_before and not part.must_startswith("/"):
                        # location /site/(.*) ~ { alias /lol$1; }
                        self.report_issue(directive, location, gixy.severity.MEDIUM)

            else:
                alias_has_slash_before = str(prev_part.value).endswith("/")

                if not alias_has_slash_before and not part.must_startswith("/"):
                    # location /site/(.*) ~ { alias /lol$1; }
                    self.report_issue(directive, location, gixy.severity.MEDIUM)

            prev_part = part

    def report_issue(self, directive, location, severity):
        self.add_issue(
            severity=severity,
            directive=directive,
        )



import fnmatch
import glob
import logging
import os
import re

from gixy.core.exceptions import InvalidConfiguration
from gixy.directives import block, directive
from gixy.parser import raw_parser
from gixy.parser.raw_parser import ParseException
from gixy.utils.text import to_native

LOG = logging.getLogger(__name__)


class NginxParser(object):
    def __init__(self, cwd="", allow_includes=True):
        self.cwd = cwd
        self.configs = {}
        self.is_dump = False
        self.allow_includes = allow_includes
        self.directives = {}
        self.parser = raw_parser.RawParser()
        self._init_directives()
        self._path_stack = None

    def parse_file(self, path, root=None, display_path=None):
        """Parse an nginx configuration file from disk.

        Args:
            path (str): Filesystem path to the nginx config to parse.
            root (Optional[block.Root]): Existing AST root to append into. If None, a new root is created.
            display_path (Optional[str]): Path to attribute to parsed nodes (used for stdin/tempfile attribution).

        Returns:
            block.Root: Parsed configuration tree.

        Raises:
            InvalidConfiguration: When parsing fails.
        """
        real_path = display_path if display_path else path
        LOG.debug("Parse file: {0}".format(real_path))
        root = self._ensure_root(root)
        try:
            parsed = self.parser.parse_path(path)
        except ParseException as e:
            # Preserve the underlying parser message and line info.
            base_msg = (
                getattr(e, "msg", None) or str(e) or "Failed to parse nginx config"
            )
            # Strip line number from error message (errors are not standardized)
            base_msg = re.sub(r":\d+$", "", base_msg)
            base_msg = re.sub(r":\d+$", "", base_msg)
            base_msg = re.sub(
                r"^(\s*)(\S)",
                lambda m: m.group(1) + m.group(2).upper(),
                base_msg,
                count=1,
            )

            # Add line number back to error message
            line = f" (line:{e.line})"
            error_msg = "{msg}{line}.".format(msg=base_msg, line=line)

            escaped_path = re.escape(path)
            ends_with_in_path_line = re.search(
                rf"in {escaped_path} \(line:\d+\)\.$", error_msg
            )
            if ends_with_in_path_line:
                error_msg = re.sub(
                    rf"in {escaped_path} \(line:(\d+)\)\.$",
                    rf"in {real_path} (line:\1).",
                    error_msg,
                )
            if not ends_with_in_path_line:
                error_msg = "{msg} in {filename}{line}.".format(
                    msg=base_msg, filename=real_path, line=line
                )

            LOG.error("{error}".format(error=error_msg))
            raise InvalidConfiguration(error_msg) from e

        current_path = display_path if display_path else path
        return self._build_tree_from_parsed(parsed, root, current_path)

    def parse_string(self, content, root=None, path_info=None):
        """Parse nginx configuration provided as a string/bytes.

        The content is written to a temporary file so that the underlying
        crossplane parser consistently receives a filesystem path (ensuring
        identical behavior to file-based parsing).

        Args:
            content (Union[str, bytes]): Nginx configuration text to parse.
            root (Optional[block.Root]): Existing AST root to append into. If None, a new root is created.
            path_info (Optional[str]): Path to attribute to parsed nodes (e.g., "<stdin>").

        Returns:
            block.Root: Parsed configuration tree.

        Raises:
            InvalidConfiguration: When parsing fails.
        """
        root = self._ensure_root(root)
        import tempfile

        data = (
            content
            if isinstance(content, (bytes, bytearray))
            else content.encode("utf-8")
        )
        tmp_filename = None
        try:
            with tempfile.NamedTemporaryFile(
                mode="wb", suffix=".conf", delete=False
            ) as tmp:
                tmp.write(data)
                tmp_filename = tmp.name
            return self.parse_file(tmp_filename, root=root, display_path=path_info)
        except InvalidConfiguration:
            raise
        finally:
            if tmp_filename:
                try:
                    os.unlink(tmp_filename)
                except Exception:
                    pass

    # Backward-compatible alias (deprecated). Prefer parse_string.
    def parse(self, content, root=None, path_info=None):
        return self.parse_string(content, root=root, path_info=path_info)

    def _ensure_root(self, root):
        """Return provided root or create a new one.

        Args:
            root (Optional[block.Root]): Existing root node or None.

        Returns:
            block.Root: Root node.
        """
        return root if root else block.Root()

    def _build_tree_from_parsed(self, parsed_block, root, current_path):
        """Finalize parsed data into the directive tree.

        Handles nginx -T dumps, manages current file attribution, and
        appends parsed directives into the provided root.

        Args:
            parsed_block (list): Parsed representation from RawParser.
            root (block.Root): Root node to append into.
            current_path (str): Current file path used for attribution.

        Returns:
            block.Root: The root containing parsed directives.
        """
        # Handle nginx -T dump format if detected (multi-file with file delimiters)
        if (
            len(parsed_block)
            and isinstance(parsed_block[0], dict)
            and parsed_block[0].get("kind") == "file_delimiter"
        ):
            LOG.debug("Switched to parse nginx configuration dump.")
            root_filename = self._prepare_dump(parsed_block)
            self.is_dump = True
            self.cwd = os.path.dirname(root_filename)
            parsed_block = self.configs[root_filename]

        # Parse into the provided root/parent context and keep attribution
        old_stack = self._path_stack
        self._path_stack = current_path
        self.parse_block(parsed_block, root)
        self._path_stack = old_stack
        return root

    def parse_block(self, parsed_block, parent):
        for node in parsed_block:
            if not isinstance(node, dict):
                continue
            parsed_type = node.get("kind")
            if parsed_type == "comment" or parsed_type is None:
                continue

            if parsed_type == "include":
                # include is handled specially
                path_info = self.path_info
                try:
                    self._resolve_include(node["args"], parent)
                except (
                    InvalidConfiguration
                ):  # We can continue after error in parsed include file, I guess.
                    pass
                finally:
                    self._path_stack = path_info

                continue

            parsed_name = node.get("name")
            parsed_line = node.get("line")
            if parsed_type == "block":
                parsed_args = [node.get("args", []), node.get("children", [])]
            elif parsed_type == "directive":
                parsed_args = node.get("args", [])
            elif parsed_type == "hash_value":
                parsed_args = node.get("args", [])
            else:
                # unknown or file_delimiter should not be here
                continue

            if (
                parent.name in ["map", "geo"] and parsed_type == "directive"
            ):  # Hack because included maps are treated as directives (bleh)
                if isinstance(parsed_args, list) and len(parsed_args) > 1:
                    parent_args = getattr(parent, "args", []) or []
                    hdr = "{} {}".format(parent.name, " ".join(parent_args)).strip()
                    error_msg = (
                        "Invalid {} entry with {} parameters: {} {{ {} {}; }};".format(
                            parent.name,
                            len(parsed_args),
                            hdr,
                            parsed_name,
                            " ".join(parsed_args),
                        )
                    )
                    LOG.warn(
                        'Failed to parse "{path_info}": {error}'.format(
                            path_info=self.path_info, error=error_msg
                        )
                    )
                    continue
                parsed_type = "hash_value"

            directive_inst = self.directive_factory(
                parsed_type, parsed_name, parsed_args
            )
            if directive_inst:
                # RawParser emits 'raw' for *_lua_block
                if parsed_type == "block" and node.get("raw") is not None:
                    try:
                        setattr(directive_inst, "raw", node.get("raw"))
                    except Exception:
                        pass
                # Set line number and file path
                directive_inst.line = parsed_line
                directive_inst.file = self.path_info
                parent.append(directive_inst)

    def directive_factory(self, parsed_type, parsed_name, parsed_args):
        klass = self._get_directive_class(parsed_type, parsed_name)
        if not klass:
            return None

        if klass.is_block:
            args = [to_native(v).strip() for v in parsed_args[0]]
            children = parsed_args[1]

            inst = klass(parsed_name, args)
            self.parse_block(children, inst)
            return inst
        else:
            args = [to_native(v).strip() for v in parsed_args]
            return klass(parsed_name, args)

    def _get_directive_class(self, parsed_type, parsed_name):
        if (
            parsed_type in self.directives
            and parsed_name in self.directives[parsed_type]
        ):
            return self.directives[parsed_type][parsed_name]
        elif parsed_type == "block":
            return block.Block
        elif parsed_type == "directive":
            return directive.Directive
        elif parsed_type == "hash_value":
            return directive.MapDirective
        elif parsed_type == "unparsed_block":
            LOG.warning(
                "Skip unparseable block '%s' from '%s'", parsed_name, self.path_info
            )
            return None
        else:
            return None

    def _init_directives(self):
        self.directives["block"] = block.get_overrides()
        self.directives["directive"] = directive.get_overrides()

    def _resolve_include(self, args, parent):
        pattern = args[0]
        #  TODO(buglloc): maybe file providers?
        if self.is_dump:
            return self._resolve_dump_include(pattern=pattern, parent=parent)
        if not self.allow_includes:
            LOG.debug("Includes are disallowed, skip: {0}".format(pattern))
            return

        return self._resolve_file_include(pattern=pattern, parent=parent)

    def _resolve_file_include(self, pattern, parent):
        path = os.path.join(self.cwd, pattern)
        exists = False
        for file_path in sorted(glob.iglob(path)):
            exists = True
            # parse the include into current context
            self.parse_file(file_path, parent)

        if not exists:
            # Align behavior with nginx: unmatched glob patterns are not warnings
            if glob.has_magic(path):
                LOG.debug(
                    "Include pattern '%s' matched no files from '%s'",
                    path,
                    self.path_info,
                )
            else:
                LOG.warning("File not found: '%s'", path)

    def _resolve_dump_include(self, pattern, parent):
        path = os.path.join(self.cwd, pattern)
        found = False
        for file_path, parsed in self.configs.items():
            if not fnmatch.fnmatch(file_path, path):
                continue
            found = True

            # Flatten includes by parsing into the current parent context.
            # We only switch the path stack for correct file attribution but keep
            # cwd unchanged (prefix-based) so relative includes inside dumps
            # resolve as they commonly do in nginx deployments.
            # This intentionally diverges from commit 0ef30ce (which switches cwd
            # to the included file directory) to avoid mis-resolving patterns like
            # sites/default.conf including conf.d/listen → /etc/nginx/conf.d/listen.
            old_stack = self._path_stack
            self._path_stack = file_path

            self.parse_block(parsed, parent)

            self._path_stack = old_stack

        if not found:
            # Align behavior with nginx: unmatched glob patterns are not warnings
            if glob.has_magic(path):
                LOG.debug("Include pattern matched no files: {0}".format(path))
            else:
                LOG.warning("File not found: {0}".format(path))

    def _prepare_dump(self, parsed_block):
        filename = ""
        root_filename = ""
        for node in parsed_block:
            if isinstance(node, dict) and node.get("kind") == "file_delimiter":
                if not filename:
                    root_filename = node.get("file")
                filename = node.get("file")
                self.configs[filename] = []
                continue
            self.configs[filename].append(node)
        return root_filename

    @property
    def path_info(self):
        """Current file being parsed, or None."""
        return self._path_stack if self._path_stack else None



from gixy.parser.raw_parser import RawParser


def test_directive_simple():
    config = """
user http;
    """
    nodes = RawParser().parse(config)
    assert len(nodes) == 1
    n = nodes[0]
    assert n.get('kind') == 'directive'
    assert n.get('name') == 'user'
    assert n.get('args') == ['http']


def test_if_args_flatten():
    config = """
if ($request_method = POST) {
}
    """
    nodes = RawParser().parse(config)
    # find the if-block
    if_nodes = [n for n in nodes if n.get('kind') == 'block' and n.get('name') == 'if']
    assert len(if_nodes) == 1
    ifn = if_nodes[0]
    assert ifn.get('args') == ["$request_method", "=", "POST"]


def test_lua_block_raw_content():
    config = """
location = /lua {
 # MIME type determined by default_type:
 default_type 'text/plain';

 content_by_lua_block {
     local res = ngx.location.capture("/some_other_location")
     if res then
         ngx.say("status: ", res.status)
         ngx.say("body:")
         ngx.print(res.body)
     end
 }
}
    """
    nodes = RawParser().parse(config)

    def find_by_name(ns, name):
        for x in ns:
            if x.get('name') == name:
                yield x
            if x.get('kind') == 'block':
                yield from find_by_name(x.get('children', []), name)

    lua_blocks = [b for b in find_by_name(nodes, 'content_by_lua_block') if b.get('kind') == 'block']
    assert len(lua_blocks) == 1
    lua = lua_blocks[0]
    assert lua.get('args') == []
    assert lua.get('children') == []
    assert isinstance(lua.get('raw'), list) and len(lua.get('raw')) == 1
    assert 'ngx.location.capture' in lua['raw'][0]


def test_file_delimiters_from_dump():
    config = """
# configuration file /etc/nginx/nginx.conf:
http {
    include sites/*.conf;
}

# configuration file /etc/nginx/sites/default.conf:
server {

}
    """
    nodes = RawParser().parse(config)
    files = [n['file'] for n in nodes if n.get('kind') == 'file_delimiter']
    assert files == ['/etc/nginx/nginx.conf', '/etc/nginx/sites/default.conf']


def test_inline_comments_filtered_but_standalone_kept():
    config = """
# Standalone
add_header X-Some-Comment some;

if (1) # Inline
{
    add_header X-Inline blank;
}
    """
    nodes = RawParser().parse(config)
    texts = [n.get('text') for n in nodes if n.get('kind') == 'comment']
    assert 'Standalone' in texts
    # Inline comment should not appear as a standalone comment node
    assert not any(t and 'Inline' in t for t in texts)





import gixy
from gixy.directives.block import GeoBlock, MapBlock
from gixy.directives.directive import MapDirective
from gixy.plugins.plugin import Plugin


class hash_without_default(Plugin):
    summary = "Missing default value in hash blocks (map/geo)."
    severity = gixy.severity.MEDIUM
    description = "A hash block without a default value may allow the bypassing of security controls."
    help_url = "https://gixy.io/plugins/hash_without_default/"
    directives = ["map", "geo"]

    def audit(self, directive):
        # Collect entries from this block, following includes when present
        if isinstance(directive, MapBlock):
            entries = list(directive.gather_map_directives(directive.children))
        elif isinstance(directive, GeoBlock):
            entries = list(directive.gather_geo_directives(directive.children))
        else:
            entries = directive.children

        found_default = False
        for child in entries:
            if (
                isinstance(child, MapDirective)
                and child.src_val == "default"
                and child.dest_val is not None
            ):
                found_default = True
                break

        if found_default:
            return

        # Special-case: for map blocks, a single mapping entry without an explicit default
        # is commonly used to intentionally yield an empty string for all other cases.
        # This is especially useful with limit_req/limit_conn where empty keys disable limits.
        # In that scenario, requiring an explicit default creates noise. Therefore, only warn
        # for map when there are two or more mapping entries and no explicit default.
        if directive.name == "map":
            entries_count = 0
            for child in entries:
                if not isinstance(child, MapDirective):
                    continue
                if child.src_val == "default":
                    continue
                entries_count += 1

            # Do not warn for single-entry maps without default
            if entries_count <= 1:
                return

        # geo continues to require an explicit default regardless of entries count
        self.add_issue(
            directive=[directive] + directive.children,
            reason="Missing `default` entry in `{0}` for variable `${1}`.".format(
                directive.name, directive.variable
            ),
        )



import gixy
from gixy.plugins.plugin import Plugin


class try_files_is_evil_too(Plugin):
    """
    Insecure example:
        location / {
            try_files $uri $uri/ /index.php$is_args$args;
        }
    """

    summary = "try_files used without open_file_cache."
    severity = gixy.severity.MEDIUM
    description = "Using try_files without open_file_cache adds extra stat() calls per request and can cause significant performance overhead."
    help_url = "https://gixy.io/plugins/try_files_is_evil_too/"
    directives = ["try_files"]

    def audit(self, directive):
        # search for open_file_cache ...; on the same or higher level
        open_file_cache = directive.find_single_directive_in_scope("open_file_cache")
        if not open_file_cache or open_file_cache.args[0] == "off":
            self.add_issue(
                severity=gixy.severity.MEDIUM,
                directive=[directive]
                + ([open_file_cache] if open_file_cache is not None else []),
                reason="`try_files` introduces extra filesystem lookups without `open_file_cache`.",
            )



import re

import gixy
from gixy.core.context import get_context
from gixy.core.variable import compile_script
from gixy.plugins.plugin import Plugin


class ssrf(Plugin):
    """
    Insecure examples:
        location ~ /proxy/(.*)/(.*)/(.*)$ {
            set $scheme $1;
            set $host $2;
            set $path $3;
            proxy_pass $scheme://$host/$path;
        }

        location /proxy/ {
            proxy_pass $arg_some;
        }
    """

    summary = "Possible SSRF (Server Side Request Forgery) vulnerability."
    severity = gixy.severity.HIGH
    description = "The configuration may allow an attacker to make arbitrary requests from the vulnerable server."
    help_url = "https://gixy.io/plugins/ssrf/"
    directives = ["proxy_pass"]

    def __init__(self, config):
        super(ssrf, self).__init__(config)
        self.parse_uri_re = re.compile(r"(?P<scheme>[^?#/)]+://)?(?P<host>[^?#/)]+)")

    def audit(self, directive):
        value = directive.args[0]
        if not value:
            return

        context = get_context()
        if context.block.name == "location" and context.block.is_internal:
            # Exclude internal locations
            return

        parsed = self.parse_uri_re.match(value)
        if not parsed:
            return

        res = self._check_script(parsed.group("scheme"), directive)
        if not res:
            self._check_script(parsed.group("host"), directive)

    def _check_script(self, script, directive):
        for var in compile_script(script):
            if var.must_contain("/"):
                # Skip variable checks
                return False
            if var.can_contain("."):
                # Yay! Our variable can contain any symbols!
                reason = "Variable `${var}` can contain untrusted user input.".format(
                    var=var.name
                )
                self.add_issue(directive=[directive] + var.providers, reason=reason)

                return True
        return False
